/******************************************************************************
 * data_red.c
 * Federico Bareilles
 * 31/01/2003
 *
 *
 *****************************************************************************/

#define _POSIX_SOURCE 1         /* POSIX compliant source  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tipper_red.h"


int data_correction(struct setup *s)
{
  struct tdata *d;
  struct tpoint *p;
  struct instrument *t = s->tipper;
  int i;
  float factor[2];
  int nf;
  float v_comp;

  /* Check for Vref goinbt to Vcc, adjust factor accordingly */
  /* Set correction factor for the reference voltage of the UNIT ADC */
  factor[0] = ((t->vref1_ADC - t->agrd_ADC)/4500.0); /* NORMAL */
  factor[1] = ((t->vref2_ADC - t->agrd_ADC)/4500.0); /* Vref=Vcc */

  d = s->data;
  while(d != NULL){
    nf = 0;
    v_comp = ((d->v_comp * 0.5)+t->agrd_ADC) * factor[nf];
    v_comp += zener_cor(v_comp,t);
    v_comp *= 2.0;
    /* If v_copm is low, use the backup vref2_ADC value */
    if ( d->v_comp < 4600.0) {
      d->error_flag |= VCOMPLOW;
      nf = 1;
      v_comp = ((d->v_comp * 0.5)+t->agrd_ADC) * factor[nf];
      v_comp += zener_cor(v_comp,t);
      v_comp *= 2.0;
    }
    d->v_comp = v_comp;

    /* Corrected voltages measured by AD converter.            */
    /* voltages are first corrected for A/D ref. voltage error */
    /* and are then corrected for the non-linear zener error   */
    d->t_adq  = (d->t_adq + t->agrd_ADC) * factor[nf];
    d->t_adq += zener_cor(d->t_adq,t);
    d->t_abs  = (d->t_abs + t->agrd_ADC) * factor[nf];
    d->t_abs += zener_cor(d->t_abs,t);
    d->t_gun  = (d->t_gun + t->agrd_ADC) * factor[nf];
    d->t_gun += zener_cor(d->t_gun,t);
    d->t_amp  = (d->t_amp + t->agrd_ADC) * factor[nf];
    d->t_amp += zener_cor(d->t_amp,t);
    d->v_comp_sigma  = d->v_comp_sigma * factor[nf];
    d->v_comp_sigma += zener_cor(d->v_comp_sigma,t);
    d->v_gun  = ((d->v_gun * 0.25)+ t->agrd_ADC) * factor[nf];
    d->v_gun += zener_cor(d->v_gun,t);
    d->v_gun *= 4.0;
    d->v_gun_sigma  = d->v_gun_sigma * factor[nf];
    d->v_gun_sigma += zener_cor(d->v_gun_sigma,t);

    for(i=0; d->n > 2 && i < d->n; i++){
      p = d->points+i;
      p->v_sky  = (p->v_sky + t->agrd_ADC) * factor[nf];
      p->v_sky += zener_cor(p->v_sky,t);
      p->v_sky_sigma  = p->v_sky_sigma * factor[nf];
      p->v_sky_sigma += zener_cor(p->v_sky_sigma,t);
      p->v_ref  = (p->v_ref + t->agrd_ADC) * factor[nf];
      p->v_ref += zener_cor(p->v_ref,t);
      p->v_ref_sigma  = p->v_ref_sigma * factor[nf];
      p->v_ref_sigma += zener_cor(p->v_ref_sigma,t);
    }
    d=d->next;
  }
  
  return 0;
}


int data_check_airmass(struct setup *s)
{
  struct tdata *d;
  int i,result=0;
  
  d=s->data;
  while(d != NULL){
    for(i=1; d->n > 2 && !(d->error_flag & OBSTACLE) && i < d->n; i++){
      if(d->points[i-1].v_sky >= d->points[i].v_sky){
	d->error_flag |= OBSTACLE;
	//PVERB("# WARNING: v_sky does not increase with Airmass at date %8.3f\n", d->date);
	result++;
      }
    }
    d=d->next;
  }
 
  return result;
}


#if 0
double sec(double x)
{
  return 1.0/cos(x);
}


float ang_to_mas(float x)
{
  return sec(x) - \
    0.0018167     * (sec(x) - 1.0) - \
    pow(0.002875  * (sec(x) - 1.0), 2) - \
    pow(0.0008083 * (sec(x) - 1.0), 3);
}
#endif


int data_reduction(struct setup *s)
{
  struct tdata *d;
  struct tpoint *p;
  int i;
  int read_ang = s->read_ang;
  float tet;
  float x[MAX_ANG],y[MAX_ANG],sigma[MAX_ANG],xx[MAX_ANG];
  float s_fac;

  d=s->data;
  while(d != NULL){
    if(d->n < read_ang) d->error_flag |= BADANGLES;
    /* Data transformation */
    for(i=0; d->n >= read_ang && i < read_ang; i++){
      p = d->points+i;

      /***************************/
      tet = (2.0 * (p->pos * 1.8 - 45.0) + 30.0) * DEG_RAD;
      //tet = acos(1.0/p->airmass);
      /***************************/
      
      /* Radiometer Beam Position on Sky (air mass) */
      x[i] = 1.0/cos(tet+(ANGOFF*DEG_RAD)+(2.0*MIROFF*DEG_RAD));
      //x[i] = ang_to_mas(tet+(ANGOFF*DEG_RAD)+(2.0*MIROFF*DEG_RAD));
      /* Radiometer Mirror Position (air mass) */
      xx[i]=1.0/cos(tet+(2.0*MIROFF*DEG_RAD));
      y[i] = p->v_ref - GAINOFF * p->v_sky;
      if( y[i] <= 0.0 ){
	y[i] = 1.0;
	d->error_flag |= SKY_G_REF;
      }
      /* OJO: /y[i] es parte del modelo... como me lo saco ??? */
      sigma[i] = s->sigma_propagation(p);
      //sigma[i] = sqrt(SQR(p->v_sky_sigma)+SQR(p->v_ref_sigma))/y[i];

      /* SPILLOVER CORRECTION FACTOR:   */
      s_fac = s->spillover(xx[i]);
      y[i] = s->y_transform(y[i]/s_fac); /********* OJO ********/
    }
    if(!(d->error_flag & (BADANGLES | SKY_G_REF)) ){
      d->f = s->fit_metod(x, y, sigma, read_ang, d->t_adq, d->t_abs);
      if(d->f != NULL){
	if( d->f->tau <= 0.0 ) d->error_flag |= BADTAU;
      }else{
	d->error_flag |= NU_ERROR;
      }
    }
    d=d->next;
  }
  
  return 0;
}

/****************************************************************************/


#if 0
int data_reduction_linear(struct setup *s)
{
  struct tdata *d;
  struct tpoint *p;
  int i;
  int read_ang = s->read_ang;
  float tet;
  float x[MAX_ANG],y[MAX_ANG],sigma[MAX_ANG],xx[MAX_ANG];
  float s_fac;

  d=s->data;
  while(d != NULL){
    if(d->n < read_ang) d->error_flag |= BADANGLES;
    /* Data transformation */
    for(i=0; d->n >= read_ang && i < read_ang; i++){
      p = d->points+i;

      tet = acos(1.0/p->airmass);      
      /* Radiometer Beam Position on Sky (air mass) */
      x[i] = -1.0/cos(tet+(ANGOFF*DEG_RAD)+(2.0*MIROFF*DEG_RAD));
      /* Radiometer Mirror Position (air mass) */
      xx[i]=1.0/cos(tet+(2.0*MIROFF*DEG_RAD));

      y[i] = p->v_ref - GAINOFF * p->v_sky;
      if( y[i] <= 0.0 ){
	y[i] = 1.0;
	d->error_flag |= SKY_G_REF;
      }
      sigma[i] = sqrt(SQR(p->v_sky_sigma)+SQR(p->v_ref_sigma))/y[i];

      /* SPILLOVER CORRECTION FACTOR:   */
      s_fac = s->spillover(xx[i]);

      y[i] = log(y[i]/s_fac);
    }
    if(!(d->error_flag & (BADANGLES | SKY_G_REF)) ){
      d->f = linear_fit(x, y, sigma, read_ang);
      if(d->f != NULL){
	if( d->f->tau <= 0.0 ) d->error_flag |= BADTAU;
      }
    }
    d=d->next;
  }
  
  return 0;
}
#endif

#if 0
int data_reduction_no_linear(struct setup *s)
{
  struct tdata *d;
  struct tpoint *p;
  int i;
  int read_ang = s->read_ang;
  float tet;
  float x[MAX_ANG],y[MAX_ANG],sigma[MAX_ANG],xx[MAX_ANG];
  float s_fac;

  d=s->data;
  while(d != NULL){
    if(d->n < read_ang) d->error_flag |= BADANGLES;
    /* Data transformation */
    for(i=0; d->n >= read_ang && i < read_ang; i++){
      p = d->points+i;

      tet = acos(1.0/p->airmass);      
      /* Radiometer Beam Position on Sky (air mass) */
      x[i] = 1.0/cos(tet+(ANGOFF*DEG_RAD)+(2.0*MIROFF*DEG_RAD));
      /* Radiometer Mirror Position (air mass) */
      xx[i]=1.0/cos(tet+(2.0*MIROFF*DEG_RAD));

      y[i] = p->v_ref - GAINOFF * p->v_sky;
      if( y[i] <= 0.0 ){
	y[i] = 1.0;
	d->error_flag |= SKY_G_REF;
      }
      sigma[i] = sqrt(SQR(p->v_sky_sigma)+SQR(p->v_ref_sigma));

      /* SPILLOVER CORRECTION FACTOR:   */
      s_fac = s->spillover(xx[i]);

      y[i] = y[i]/s_fac;/********* OJO ********/
    }
    if(!(d->error_flag & (BADANGLES | SKY_G_REF)) ){
      d->f = no_linear_fit(x, y, sigma, read_ang);
      if(d->f != NULL){
	if( d->f->tau <= 0.0 ) d->error_flag |= BADTAU;
      }
    }
    d=d->next;
  }
  
  return 0;
}
#endif



/***************************************************************************
 * Funcion original de los mejicanos:
 ***************************************************************************/
#if 0
void fit(x,y,ndata,sig,a,b,siga,sigb,chi2,q)
float x[],y[],sig[],*a,*b,*siga,*sigb,*chi2,*q;
int ndata;
{
  int i;
  float wt,t,sxoss,sx=0.0,sy=0.0,st2=0.0,ss;
  //float gammq();
  
  *b=0.0; /*  y = b * x + a  */
  ss=0.0;
  for (i=1;i<=ndata;i++) {
    wt=1.0/SQR(sig[i]);
    ss += wt;
    sx += x[i]*wt;
    sy += y[i]*wt;
  }
  sxoss=sx/ss;
  for (i=1;i<=ndata;i++) {
    t=(x[i]-sxoss)/sig[i];
    st2 += t*t;
    *b += t*y[i]/sig[i];
  }
  *b /= st2;
  *a=(sy-sx*(*b))/ss;
  *siga=sqrt((1.0+sx*sx/(ss*st2))/ss);
  *sigb=sqrt(1.0/st2);
  *chi2=0.0;
  for (i=1;i<=ndata;i++)
    *chi2 += SQR((y[i]-(*a)-(*b)*x[i])/sig[i]);
  //*q=gammq(0.5*(ndata-2),0.5*(*chi2));
  
}

#endif




